######################################################################
# vsqtxt2lipsync_041.pl : 
# 
# extract information for lip-sync  out of vsq file.
#
# author: zhuo   October-November 2007
######################################################################
#
# usage:
# % mf2t a.vsq  | perl vsqtxt2lipsync_041.pl [-t track_spec] [-p phoneme_table] [-m (e0|e1|e2|f|t|v)] [xx.xx [yy.yy]]  > lipsync.txt
#
#  -t track_spec : which track to output:
#      0      ...  track[0]
#      1,3,4  ...  track[1],track[3], track[4].
#      all    .... all the tracks.
#  -p phoneme_table : text file name that describes associations between
#      a phoneme and the symbol.
#      Note that this works with -m f or -m e2  option only.
#  -m e0: event-dump, always put "end" for "beg"
#  -m e1: event-dump, put "sil" only when a syllable ends early
#  -m e2: event-base, put "sil" only when a syllable ends early,
#      with RICHIER INFORMATION.  has control lines for easier
#      file loading.
#  -m f : frame-based dump.
#  -m t : print track numbers and names and exit, doing nothing.
#  -m v : print raw text in metaevents.
#
#  xx.xx : frames per second.  can be fractional.  default: 30.00
#  yy.yy : initial-frame-offset, in millisecond. can be fractional. default:0.00
#
#
# Rev. 0.41 Nov. 23, 2007
#   - -m e2 option added. 
#   - -m t  option : just printout track names and exit.
#   - old -t option is now  replaced with as "-p", phoneme-table option.
#   - bug-fix in vsq text handling. '\\' remained.  conversion added.
#
# Rev. 0.40 Nov. 3, 2007
#   - xx.xx, yy.yy could not be read. fixed.
#   - -m (e0|e1|f) introduced
#
# Rev. 0.30 Oct 29, 2007
#   frame-wise output, the first version... simply the 
#   vowels for each frame is picked up.
#
# Rev. 0.20 Oct 20, 2007
#   index-to-time conversion implemented.
#
# Rev.10 Oct 20, 2007
#   extract text meta events, merge them and chop at \n.
#
#
$vocalips_revision = "0.41";


######################################################################
## default global params.
######################################################################
## smf_type 
##    0: all in one track, 1:track-by-track
## smf_ntracks
##    must be 1 or above
## smf_tpq
##    tick-per-quarter.
##    One quarter note is represented by index interval 480 by default.
## smf_eos_tick 
##    the tick of the moment of eos (end of sequence).
##    the max value of all the tracks.
## smf_eos_usec
##    the time of the moment of eos (end of sequene).
##    = idx_to_time( $smf_eos_tick)
$smf_type = 0;       
$smf_ntracks = 0;    
$smf_tpq = 480;
$smf_eos_tick = 0;
$smf_eos_usec = 0.0;

######################################################################
## tempo and timesig 
######################################################################
## We assume that tempo-change / timesig-change will appear in a
## single master-track only, thus they are in the ascending order.
## Therefore sorting is not applied (so far).

## smf_tempo_list.
## this would look like: ( [idx,tempo,time], [idx,tempo,time], ... )
## to keep the tempo curve.
## Note that this tempo value is "time length of one quarter note in
## microsecond";
## e.x.  666667 microsecond <=> 0.666667 sec for one beat <=> BPM=90.
@smf_tempo_list = ();

## smf_timesig_list.
## this  would look like: ( [idx,meter,num,num], [idx,meter,num,num], .. )
## to keep the timesig markers.
@smf_timesig_list = ();


@output_phoneme_match_table = ();
#    ['SIL', "M"],
#    ['a$', "A"],
#    ['i$', "I"],
#    ['M$', "U"],
#    ['e$', "E"],
#    ['o$', "O"],
#    ['.*', "N"]
#);


######################################################################
# read commandline options
######################################################################

## default output params: 
## fps (frame-per-second)  in msec.
## ifo (initial frame offset) in msec.
## output_mode: (-m) set to "e2".
## output track spec . (-t) "all"

$output_fps = 30.0000;
$output_ifo_msec =  0.0000;
$output_mode = "e2";
$output_track_spec = "all";


foreach $arg (@ARGV) {
    $cmdline .= "\t$arg";
}

while ( 1 ) {
    my $arg = $ARGV[0]; 
    unless ( $arg =~ /^\-/ ) { last; }

    shift;

    if ( $arg eq "-t" ) {
        $output_track_spec = $ARGV[0];
        shift;
        if ( $output_track_spec eq "" ) { die "track_spec not specified. \n"; }
        next;
    }
    if ( $arg eq "-p" ) {
        my $table_file = $ARGV[0];
        shift;
        @output_phoneme_match_table = ();
        open( TABLE, $table_file ) || die "can't open table file\n";
        while ( <TABLE> ) {
	    chop;
	    s/\#.*$//g;
            if ( /^\s*$/ ) {
	        next;
            }
	    my ($exp, $pat) = split( /\s+/ );
	    push( @output_phoneme_match_table, [$exp, $pat] );
#DP            print "exp=$exp, pat=$pat: recorded on phneme_mach_table\n";
        }
        close TABLE;
        next;
    }
    if ( $arg eq "-m" ) {
	$output_mode = $ARGV[0];
        shift;
        if ( ! ($output_mode =~ /^(e0|e1|e2|f|t|v)$/) ) {
            die "undefined mode $output_mode specified for -m option\n";
        }
        next;
    }
    die "undefined option $arg specified.";
}

if ( $ARGV[0] ne "" ) {
    $output_fps = $ARGV[0];
    shift; 
    if ( $ARGV[0] ne "" ) {
	$output_ifo_msec = $ARGV[0];
	shift;
    }
}


######################################################################
## load track-wise data
######################################################################
$smf_current_track_idx = -1;


########################################
## read thru the vsq file which was converted to text by mf2t in advance
########################################

while (<>) {
    chop;
    $line = $_;     ### for later use
    @words = split / /;
    if ($words[0] eq "MFile") {
	($dummy, $smf_type, $smf_ntracks, $smf_tpq) = @words;
#	print "smf_type = $smf_type, ";
#	print "smf_ntracks = $smf_ntracks, ";
#	print "smf_tpq = $smf_tpq\n" ;
	next;
    }

    # tempo information is collected from all the tracks
    if ($words[1] eq "Tempo") {
	push( @smf_tempo_list, [$words[0], $words[2], 0.0] );
	next;
    }
    if ($words[2] eq "TimeSig") {
	push( @smf_timesig_list, [$words[0], $words[2], $words[3], $words[4]] );
	next;
    }

    if ($words[1] eq "Meta" && $words[2] eq "TrkName") {
	++$smf_current_track_idx;
	my $text = $line;
	$text =~ s/^[0-9]+\s+Meta\s+TrkName\s+//;
	$text =~ s/^\"//;
	$text =~ s/\"$//;
	$smf_track_name[$smf_current_track_idx] = $text;
    }
    if ($smf_current_track_idx < 0) { next; }

    ### the following section is visited only after 
    ### the first "TrkName" is reached (that is, smf_current_track_idx
    ### has been set to 0)

    if ( ($words[1] eq "Meta") && ($words[2] eq "Text") && ($words[3] =~ /^\"DM/)) {
	$line =~ s/^[0-9]+\s+Meta\s+Text\s+//;
	$line =~ s/^\"DM:[0-9]{4}://;
	$line =~ s/\"$//;
	$smf_metatext_dm[$smf_current_track_idx] .= $line;
	next;
    }

    if ( ($words[1] eq "Meta") && ($words[2] eq "TrkEnd") ) {
	if ( $smf_eos_tick < $words[0] ) {
	    $smf_eos_tick = $words[0];
	}
    }
}
$vsq_ntracks = $smf_current_track_idx;


########################################
## post processing : generate map for converting tick-index to time
########################################
$time_now_usec = 0.0;
$prev_idx = 0;
$prev_tempo = 500000;
foreach $pIdxTempo ( @smf_tempo_list ) {
    my ($idx, $tempo) = @$pIdxTempo;
    $time_now_usec += $prev_tempo * ($idx - $prev_idx) / $smf_tpq;
    $prev_idx = $idx;
    $prev_tempo = $tempo;
    $pIdxTempo->[2] = $time_now_usec;
#    print "$pIdxTempo->[0]   $pIdxTempo->[1]   $pIdxTempo->[2]\n";
}    
## after this processing, smf_tempo_list looks like:
##   ( [idx,tempo, time], [idx,tempo, time], [idx,tempo, time].... )
## here, time means the elapsed time from the beginning of the track
## at the moment of index.

########################################
## conversion func. from sequence index to absolute time (in micro-sec).
## Note: this can be used only after building smf_tempo_list.
########################################
## 
## Note: you might think that binary-search would make things faster;
## However, given the fact that the number of tempo-changes are at
## maximum 10 or so, simple linear-search would not be so bad.

sub idx_to_time {
    my ( $target ) = @_;
    my ( $pIdxTempo, $idx, $tempo, $timeoffset );
    foreach $pIdxTempo (reverse @smf_tempo_list) {
#	printf( "target=%d,  table=%d\n", $target, $pIdxTempo->[0] );
	if ( $target >= ($pIdxTempo->[0]) ) {
	    ($idx, $tempo, $timeoffset) = @$pIdxTempo;
	    last;
	}
    }
#    printf("point chosen: idx=%d, tempo=%d, time=%f\n", $idx, $tempo, $timeoffset );
    return (0.0 + $target - $idx) * $tempo / $smf_tpq + $timeoffset;
}

## testing code for idx_to_time.
## I had manually inserted the following lines to miku_vsq.txt, after
## the line that said "Tempo 666667" for test.
##700 Tempo 500000
##1000 Tempo 800000
##2000 Tempo 750000
##3000 Tempo 500000
##for $i ( 0, 1, 500, 999, 1000, 1001, 1199, 1200, 1999, 2000, 2001 ) {
##    $t = idx_to_time($i);
##    printf( "%5d => %f\n", $i, $t );
##}


########################################
### fix merged textevents  and store it in $smf_metatext_dm[trk].
########################################
$smf_eos_usec = idx_to_time( $smf_eos_tick );


######################################################################
# summary of the processing above:
#   - ticks per quarter in:
#     $smf_tpq
#   - the number of the tracks in: 
#     $vsq_ntracks = $smf_current_track_idx;
#     (note that an smf track may have no vsq events - as in 
#     the master tempo track. so vsq_ntracks < smf_ntracks.)
#   - track names in:
#     $smf_track_name[ $vsq_track_index ];
#   - the time length of the longest track in:
#     $smf_eos_usec
#   - a string of merged and formatted text meta events  
#     (for each track) in:
#     $smf_metatext_dm[ $track_index ]
#   - tick-to-time mapping by scanning tempo metaevent
#     in all tracks --- now user can obtain the elapsed time
#     by:     idx_to_time( $tick_index );
#     which takes all the tempo changes into consideration.
######################################################################


######################################################################
# for tracks specified, do:
#   - text formatting of vsq info
#   - parse the formatted vsq info into the internal data structure
#   - print out the necessary information by scanning the data structure
######################################################################

if ( $output_track_spec eq "all" ) {
    @output_track_indice = (0..$vsq_ntracks);
} else {
    @output_track_indice = split( /,/, $output_track_spec );
}

if ( $output_mode eq "e2" ) {
    printf( "##\tbegin\tvocalips\t%s%s\n", $vocalips_revision, $cmdline );
}

foreach $i ( @output_track_indice ) {
    if ( $i < 0 || $i > $vsq_ntracks ) {next;}
    my $vsq_text = $smf_metatext_dm[$i];
    $vsq_text =~ s/\\n/\n/g;
    $vsq_text =~ s/\\\"/\"/g;
    $vsq_text =~ s/\\\\/\\/g;   # fu -> p\ M , not p\\ M. fixed in Rev. 0.41
    $smf_vsq_text[$i] = $vsq_text; # save a string of all the text

## read in the text into a structure
    my $ref = vsq_text_to_structure( $vsq_text );

##    test dump of the structure
    if ( 0 ) {
	vsq_structure_to_text( $ref );
    }

## generates syllable_anchors data from the internal structure 
    my $syllable_anchors = vsq_structure_to_syllable_anchors( $ref );

    if ( $output_mode eq "t" ) {
	printf("%3d:%s\n", $i, $smf_track_name[$i] );
	next;
    }

    if ( $output_mode eq "e2" ) {
	printf( "##\tbegin\ttrack\t%s\n", $smf_track_name[$i] );
	printf( "##\tbegin\teventlist\n" );
    } else {
	printf( "***(%s)***\n", $smf_track_name[$i] );
    }

    if ( $output_mode eq "v" ) {
      print $vsq_text;
      next;
    }

    if ( $output_mode =~ /^e/ ) {
        ## Rev.0.20, or Rev.0.21 mode ... event-dump
        ## this function has two modes - one is Rev.0.20 compatible mode, 
        ## the other is Rev.0.21 and later, for improved listing 
        ##  - "sil" is not listed if overridden by the next syllable;
        ##  - if syllable-end comes early, "sil" is put, with "next syllable"
        ##    announcement .... which is necessary for shut-consonants
	syllable_anchors_print( $syllable_anchors );
    } else {
	## Rev.0.30 mode ... frame-dump
	my $frames = syllable_anchors_to_frames_print( $syllable_anchors, $output_fps, $output_ifo_msec );
    }
    if ( $output_mode eq "e2" ) {
	printf( "##\tend\teventlist\n" );
	print "##\tend\ttrack\n";
    }
}

if ( $output_mode eq "e2" ) {
    printf( "##\tend\tvocalips\n" );
}





######################################################################
## read-in vsq_text  and store it into a structure, of which reference
## is returned.
##
## pRef= vsq_text_to_structure($text);
##
## print $pRef->{"Common"}->{"Version"};
## .... this will print "DSB301", or something like that.
## 
## section "EventList", which associates tick indice with ID#nnnn, 
## is stored as "timevalue" type, which can be accessed as
## print $pRef->{"EventList"}->[0]->[0]
## .... the first tick index .
## print $pRef->{"EventList"}->[0]->[1]
## .... "ID#nnnn"  for the first tick index.
## 
## section "*BPList", which associates tick indice to a certain value,
## is stored as "timevalue" type, which can be accessed as
## print $pRef->{"XXXBPList"}->[0]->[0]
## .... the first tick index.
## print $pRef->{"XXXBPList"}->[0]->[1]
## .... number  for the first tick index.
##
## To know if the set is a timevalue or an array, 
## the following is used.
## if ( $pRef->{"isTimevalue"}->{$section} == 1 ) { ...
##
## The order of the sections is preserved in a list at:
##  $pRef->{"sectionOrder"}

sub vsq_text_to_structure {
    my ($vsq_text) = @_;
    my $pVsqStructure = {};
    my ( $section, $pNewSet, $isTimevalue );
    chop $vsq_text;
    my $pIsTimevalue = {};
    my $pSectionOrder = [];
    $pVsqStructure->{"isTimevalue"} = $pIsTimevalue;
    $pVsqStructure->{"sectionOrder"} = $pSectionOrder;
    foreach $line (split( /\n/, $vsq_text )) {
	if ($line =~ /^\[(.+)\]$/ ) {
	    $section=$1;
	    if ($section =~ /BPList$/ ||$section eq "EventList" ) {
		$pNewSet = [];
		$isTimevalue = 1;
	    } else {
		$pNewSet = {};
		$isTimevalue = 0;
	    }
	    $pVsqStructure->{$section} = $pNewSet;
	    $pIsTimevalue->{$section} = $isTimevalue;
	    push( @$pSectionOrder, $section );
	    next;
	}

	$line =~ /(^[^=]+)=(.*)$/;
	if ( $isTimevalue ) {
	    push (@$pNewSet, [$1, $2] );
	} else {
	    $pNewSet->{$1} = $2;
	}
    }
    return $pVsqStructure;
}


######################################################################
## print vsq_structure
######################################################################
sub vsq_structure_to_text {
    my ($ref) = @_;

    my ($pair, $pSection, $section, $key);
    
    my $pSectionList = $ref->{"sectionOrder"};

    foreach $section (@$pSectionList) {
	printf ("[%s]\n", $section );
	$pSection = $ref->{$section};
	if ( $ref->{"isTimevalue"}->{$section} == 1 ) {
	    foreach $pair ( @$pSection ) {
		printf("%d=%s\n", $pair->[0], $pair->[1] );
	    }
	} else {
	    foreach $key (keys %$pSection) {
		printf("%s=%s\n", $key, $pSection->{$key} );
	    }
	}
    }
}
######################################################################
## generate syllable_anchors from vsq_structure.
## The return value looks like, for e0, e1, and f:
##
##  [
##    [time_usec, beg/sil, chars, phonetic, dynamic, note_num, note_name],
##    [time_usec, beg/sil, chars, phonetic, dynamic, note_num, note_name],
##    :
##  ]
## The return value looks like, for e2, 
##  [
##    [start_usec, next_start_usec, UTT/SIL, chars, phonetic_0, phonetic_1, lipclass_0, lipclass_1, dyn, note_num ]
##    :
##  ]
##  start_usec:  time-index, from the beginning of the track
##  next_start_usec: time-index, from the beginning of the track, of the next line
##               if this is the last line, then next_start_usc = start_usec.
##  UTT/SIL:     UTT is for utterance, SIL for Silence.
##  chars:       katakana-s.
##  phonetic_0:  vocaloid-phonetic. initial consonant part, if any. 'n' is here, too.
##  phonetic_1:  vocaloid-phonetic. vowel part, if any. 
##  dyn:         dynamics
##  lipclass:    lip-class0:  m(,b,p), s(,z,ch,dz,..), n(,tdr), k(,g) y, w, .(h)
##                 m : lips touch
##                 s : lips open, teeth touch
##                 n : lips open, tongue touch
##                 k : lips open, inside narrows
##                 y : lips like vowel "i"
##                 w : lips like vowel "u"
##                 . : none. 
##               lip-class1: aiueo. (.: keep class0.)

######################################################################



sub vsq_structure_to_syllable_anchors {
    my ($ref) = @_;
    my $pEventList = $ref->{"EventList"};  # this is a timevalue set.
    my $pPair;
    my $pSyllable_anchors = [];

    my $end_time;
    my $prev_end_time = -1.0;
    
    if ($output_mode eq "e2" ) {
	foreach $pPair (@$pEventList) {
	    my ($begin_tick_index, $idnum) = @$pPair;
	    if ( $ref->{$idnum}->{"Type"} ne "Anote" ) { next; }
	    
	    my $end_tick_index = $begin_tick_index + $ref->{$idnum}->{"Length"};
	    my $dynamics = $ref->{$idnum}->{"Dynamics"};
	    my $notenum = $ref->{$idnum}->{"Note#"};
	    my $lyricHandle = $ref->{$idnum}->{"LyricHandle"};
	    my ($chars_hex, $phonetic, $v0, $v1, $v2, $v3) = split( /,/, $ref->{$lyricHandle}->{"L0"} );
	    $chars_hex =~ s/[\"\\x]//g;
	    my $chars = pack "H*", $chars_hex;
	    $phonetic =~ s/\"//g;
	    my ($phonetic_0, $phonetic_1 ) = split( / +/, $phonetic );
	    if ( $phonetic_1 eq "" ) {
		if ( $phonetic_0 =~ /[aiMeo]/ ) {
		    $phonetic_1 = $phonetic_0;
		    $phonetic_0 = "";
		}
 	    }
	    my $beg_time = &idx_to_time($begin_tick_index)/1000000.0;
	    $end_time = &idx_to_time($end_tick_index)/1000000.0;
	
	    if ( $prev_end_time > 0.0 ) {
		if ( $prev_end_time < $beg_time ) {
		    my $label = phonetic_to_matched_pattern( "SIL" );
		    push( @$pSyllable_anchors, [$prev_end_time, "SIL", $dynamics, $notenum, $chars, $phonetic_0, $phonetic_1, $label] );
		}
	    }
	    $prev_end_time = $end_time;
	    my $label = phonetic_to_matched_pattern( $phonetic );
	    push( @$pSyllable_anchors, [$beg_time, "UTT", $dynamics, $notenum, $chars, $phonetic_0, $phonetic_1, $label] );
	}
	push( @$pSyllable_anchors, [ $prev_end_time, "SIL" ] );

    } else { # e0, e1, f
	foreach $pPair (@$pEventList) {
	    my ($begin_tick_index, $idnum) = @$pPair;
	    if ( $ref->{$idnum}->{"Type"} ne "Anote" ) { next; }
	    
	    my $end_tick_index = $begin_tick_index + $ref->{$idnum}->{"Length"};
	    my $dynamics = $ref->{$idnum}->{"Dynamics"};
	    my $notenum = $ref->{$idnum}->{"Note#"};
	    my $notename = &note_number_to_name( $notenum );
	    my $lyricHandle = $ref->{$idnum}->{"LyricHandle"};
	    my ($chars_hex, $phonetic, $v0, $v1, $v2, $v3) = split( /,/, $ref->{$lyricHandle}->{"L0"} );
	    $chars_hex =~ s/[\"\\x]//g;
	    my $chars = pack "H*", $chars_hex;
	    $phonetic =~ s/\"//g;
	    my $beg_time = &idx_to_time($begin_tick_index)/1000000.0;
	    $end_time = &idx_to_time($end_tick_index)/1000000.0;
	
	    if ( $prev_end_time >= 0.0 ) {
		if ( $output_mode eq "e0" ) {
		    ## e0: Rev.0.20 mode:  always put an end for each syllable
		    if ( $prev_end_time > $beg_time ) {
			$prev_end_time = $beg_time;
		    }
		    if ( $beg_time > 0.0 ) {
			push( @$pSyllable_anchors, [ $prev_end_time, "end" ] );
		    }
		} else {
		    ## e1 or f: Rev.0.21 and later.
		    ## put an end only when a syllable ends early, leaving a silence;
		    ## also, next syllable is announced.
		    if ( $prev_end_time < $beg_time ) {
			push( @$pSyllable_anchors, [ $prev_end_time, "sil", $chars, $phonetic ] );
		    }	    
		}
	    }
	    $prev_end_time = $end_time;
	    push( @$pSyllable_anchors, [$beg_time, "beg", $chars, $phonetic, $dynamics, $notenum, $notename ] );
	}
	if ( $output_mode eq "e0" ) {
	    push( @$pSyllable_anchors, [ $end_time, "end" ] );
	} else {
	    push( @$pSyllable_anchors, [ $end_time, "sil" ] );
	}
    }
    return $pSyllable_anchors;
}


######################################################################
## print the contents of syllable_anchors.
######################################################################
sub syllable_anchors_print {
    my ( $pSylAncs ) = @_;
    my ( $pSylAncItem );

    my ( $prev_start ) = -1.0;

    if ( $output_mode eq "e2" ) {
	my $next_start = 0;
	my $max_idx = $#$pSylAncs;
	my ( $idx ) = 1;
	foreach $pSylAncItem (@$pSylAncs ) {
	    my ($start, $uttsil, $dynamics, $notenum, $chars, $phonetic_0, $phonetic_1, $label ) = @$pSylAncItem;
	    if ( $idx <= $max_idx ) {
		$next_start = $pSylAncs->[$idx]->[0];
	    } else {
		$next_start = $start;
	    }
	    printf( "%10.3f\t%10.3f\t%s\t%d\t%d\t%s\t%s\t%s\t%s\t%s\n", $start, $next_start, $uttsil, $dynamics, $notenum, $chars, $phonetic_0, $phonetic_1, $label );
	    ++$idx;
	}

    } else {  #e0, e1, f
	foreach $pSylAncItem (@$pSylAncs ) {
	    my ($t, $begend, $chars, $phonetic, $dynamics, $notenum, $notename ) = @$pSylAncItem;
	    if ( $begend eq "beg" ) {
		printf( "%10.3f\t%s\t%s\t%s\t%d\t%d\t%s\n", $t, $begend, $chars, $phonetic, $dynamics, $notenum, $notename );
	    } else {
		printf( "%10.3f\t%s\t%s\t%s\n", $t, $begend, $chars, $phonetic );
	    }		
	}
    }
}




######################################################################
## generate frame-wise vowel info  from syllable_anchors
## this is the output function for Rev.0.30.
## so far, prints directly.
## 
######################################################################

sub syllable_anchors_to_frames_print {
    my ($pSylAnc, $fps, $ifo ) = @_;

    my $usec_pf  = 1000000.0 / $fps; 
    my $max_idx = $#$pSylAnc;

    my $now_usec;
    my $sylAncNextIdx = 0;
    my $frame_cnt = 0;
    my $phonetic = "SIL";
    my $pattern = phonetic_to_matched_pattern( $phonetic );

    while (1) {
	$now_usec = $usec_pf * $frame_cnt + $ifo * 1000.0;
	if ($now_usec > $smf_eos_usec ) {
	    return;
	}
	if ( $now_usec >= ($pSylAnc->[$sylAncNextIdx]->[0]*1000000.0) ) {
	    if ( $pSylAnc->[$sylAncNextIdx]->[1] eq "beg" ) {
		$phonetic = $pSylAnc->[$sylAncNextIdx]->[3];
	    } else {
		$phonetic = "SIL";   # tentative.
	    }
	    if ( $sylAncNextIdx < $max_idx ) {
		++$sylAncNextIdx;
	    }
	    $pattern = phonetic_to_matched_pattern( $phonetic );
	}
	printf( "%5d\t%s\n", $frame_cnt, $pattern );
	++$frame_cnt;
    }
}


######################################################################
##  pattern_match table conversion
##  input syllable is matched against the regexps in '%output_phoneme_match_table'.
######################################################################
sub phonetic_to_matched_pattern {
    my ( $phonetic ) = @_;
    my $pPair;
    foreach $pPair (@output_phoneme_match_table) {
	if ( $phonetic =~ /$pPair->[0]/ ) {
#	    print "*** phonetic=$phonetic, pair=$pPair->[0], $pPair->[1]; \n";
	    return $pPair->[1];
	}
    }
    return $pPair->[1];
}

######################################################################
## utility: convert notenumber to note-name
######################################################################
# 23->-1, 24->0 35->0 36->1   59->2, 60->3, 61->3, 71->3, 72->4
sub note_number_to_name {
    my ($num) = @_;
    my @pitchclass = ("C","C#", "D", "D#","E", "F","F#","G","G#","A","A#","B");
    my ($oct) = int(($num-24) / 12 );
    my ($name) = $pitchclass[$num % 12];
    return sprintf("%-2s %2d", $name,$oct);
}    
